import torch
import torch.nn.functional as F

import os
import zipfile
from urllib import request

# adapted from https://github.com/vsitzmann/light-field-networks/blob/master/geometry.py
# label_masks, world_cams, cam_ks, n_rays=-1, depth=None
def sample_rays(world_cam, cam_k, n_rays=-1, semseg=None, depth=None, H=None, W=None, normalize=True):
    assert semseg is not None or depth is not None or (H is not None and W is not None)
    device = world_cam.device
    # get ray origins from world_cam
    ray_origs = world_cam[..., :3, 3]
    # get relevant data sizes
    batch_size, n_cams, *_ = world_cam.shape
    if semseg is not None:
        H, W = semseg.shape[-2:]
    elif depth is not None:
        H, W = depth.shape[-2:]
    else:
        H, W = H, W
    # get pixel indices
    yx = torch.cartesian_prod(torch.arange(H), torch.arange(W)).to(device)
    # parse intrinsics matrices
    fx = cam_k[..., 0, :1]
    fy = cam_k[..., 1, 1:2]
    cx = cam_k[..., 0, 2:3]
    cy = cam_k[..., 1, 2:3]
    # if desired sample random rays per camera
    if n_rays == -1:
        y_cam = yx[..., 0]
        x_cam = yx[..., 1]
    else:
        # TODO - IMPROVE UPON randperm!!!
        rand_inds = torch.randperm(H*W, device=device)[:n_rays]
        y_cam = yx[rand_inds, 0]
        x_cam = yx[rand_inds, 1]
    # obtain corresponding pixel labels if necessary
    gt_labels = semseg[...,y_cam, x_cam].permute(0,1,3,2) if semseg is not None else None
    # if necessary obtain depth
    gt_depth = depth[...,y_cam, x_cam].permute(0,1,3,2) if depth is not None else None
    # get homogeneous pixel coordinates
    x_lift = (x_cam + 0.5 - cx) / fx
    y_lift = (y_cam + 0.5 - cy) / fy
    cam_coords_hom = torch.stack([x_lift, y_lift, torch.ones_like(x_lift), torch.ones_like(x_lift)], dim=-1)
    # convert to world coordinates
    # Sitzmann et al. use this
    # -> world_coords = torch.einsum('b...ij,b...kj->b...ki', cam_pose, cam_coords_hom)[..., :3]
    # more readable version (swap if this is bad for performance)
    world_coords = (world_cam.unsqueeze(-3) @ cam_coords_hom.unsqueeze(-1))[...,:3, 0]
    # get normalized ray directions
    if normalize:
        ray_dirs = F.normalize(world_coords - ray_origs.unsqueeze(-2), dim=-1)
    else:
        ray_dirs = world_coords - ray_origs.unsqueeze(-2)
    return ray_origs, ray_dirs, gt_labels, gt_depth

def download_example_data(out_dir, verbose=True):
    data_url = 'https://uni-bonn.sciebo.de/s/EM9iewOyFuwu2sk/download'
    local_file = 'grid_fusion_pytorch_example_data.zip'
    if os.path.isdir(out_dir):
        if verbose:
            print('Output directory already exists. Check if data is already downloaded, else delete ' + out_dir + ' and try again.')
    else:
        if verbose:
            print('Downloading data.')
        request.urlretrieve(data_url, local_file)
        os.mkdir(out_dir)
        if verbose:
                print('Unpacking data archive.')
        with zipfile.ZipFile(local_file, 'r') as zip_ref:
            zip_ref.extractall(out_dir)
        print('Removing data archive after unpacking.')
        os.remove(local_file)
        
# semseg: batch_size x n_cams x h * w x 1
# out: batch_size x n_cams x h * w x C
def soften_semseg(semseg, eps=1e-4, num_classes=None):
    if num_classes is None:
        num_classes = semseg.max().item() + 1
    semseg_delta = torch.eye(num_classes).to(semseg.device)[semseg]
    # finally soften delta peaks
    ret = torch.where(semseg_delta == 0, torch.ones_like(semseg_delta)*eps, torch.ones_like(semseg_delta)-(num_classes-1)*eps)
    return ret

def calc_reflect_prob(hits, misses):
    return torch.where(hits > 0, hits/(hits + misses), 0)


import os
import torch
from torch.utils.cpp_extension import load

parent_dir = os.path.dirname(os.path.abspath(__file__))

point_cloud_fusion_util_cuda = load(
        name='point_cloud_fusion_util_cuda',
        sources=[
            os.path.join(parent_dir, path)
            for path in ['cuda/point_cloud_fusion.cpp', 'cuda/point_cloud_fusion.cu']],
        verbose=True)

def apply_point_cloud_fusion(grid_counter, point_cloud_locations, min_range, max_range, grid_semantic=None, point_cloud_logprobs=None, verbose=False, assert_inputs=True):
    if assert_inputs:
        assert grid_counter.is_cuda
        assert point_cloud_locations.is_cuda
        assert min_range.is_cuda
        assert max_range.is_cuda
        assert grid_semantic is None or (grid_semantic.is_cuda and not (grid_semantic >= 0.).any())
        assert point_cloud_logprobs is None or (point_cloud_logprobs.is_cuda and not (point_cloud_logprobs >= 0.).any())
    # TODO: assert validity of inputs and document function
    # make sure the shape of grid_counter is suitable
    if len(grid_counter.shape) == 4:
        if verbose:
            print('Warning: grid_counter should be 5D, but is 4D. Adding dummy batch dimension.')
        grid_counter = grid_counter.unsqueeze(0)
    elif len(grid_counter.shape) == 3:
        if verbose:
            print('Warning: grid_counter should be 5D, but is 3D. Adding dummy batch and channel dimensions.')
        grid_counter = grid_counter.unsqueeze(0).unsqueeze(1)
    # check how many channels the counter grid has
    n_counters = grid_counter.shape[1]
    if n_counters > 1:
        if verbose:
            print('Warning: grid_counter should have 1 channel, but has more. Only using first channel.')
    # make sure the shapes of min_range and max_range are suitable
    if len(min_range.shape) == 1:
        if verbose:
            print('Warning: min_range should be 2D, but is not. Adding dummy batch dimension.')
        min_range = min_range.unsqueeze(0)
    if min_range.shape[0] != grid_counter.shape[0]:
        if verbose:
            print('Warning: min_range should have the same batch dimension as grid_counter, but does not. Assuming equal ranges for all scenes in batch.')
        min_range = min_range.expand(grid_counter.shape[0], -1)
    if len(max_range.shape) == 1:
        if verbose:
            print('Warning: max_range should be 2D, but is not. Adding dummy batch dimension.')
        max_range = max_range.unsqueeze(0)
    if max_range.shape[0] != grid_counter.shape[0]:
        if verbose:
            print('Warning: max_range should have the same batch dimension as grid_counter, but does not. Assuming equal ranges for all scenes in batch.')
        max_range = max_range.expand(grid_counter.shape[0], -1)
    # if provided, make sure the shape of the semantic voxel grid is suitable
    if grid_semantic is not None:
        if len(grid_semantic.shape) == 4:
            if verbose:
                print('Warning: grid_semantic should be 5D, but is 4D. Adding dummy batch dimension.')
            grid_semantic = grid_semantic.unsqueeze(0)
        # compute number of class channels
        n_classes = grid_semantic.shape[1]
    else:
        n_classes = 0
    # check whether a semantic map was provided
    if n_classes > 0:
        # if there is no semantic segmentation for the point cloud, warn the user and apply counting model without it
        if point_cloud_logprobs is None:
            if verbose:
                print('Warning: semantic voxel grid detected, but no semantic segmentation provided for point_clouds! Counting hits without Bayes filter.')
            return point_cloud_fusion_util_cuda.point_cloud_hit_counter_free_function(grid_counter, point_cloud_locations, min_range, max_range)
        else:
        # if there is a semantic segmentation, make sure that the number of classes matches
            assert point_cloud_logprobs.shape[-1] == n_classes
            # modify distances for background class rays
            grid_counter_out, grid_semantic_out = point_cloud_fusion_util_cuda.point_cloud_hit_counter_bayes_free_function(grid_counter, grid_semantic, point_cloud_locations, point_cloud_logprobs, min_range, max_range)
            # normalize semantic map
            grid_semantic_out -= torch.logsumexp(grid_semantic_out, dim=1, keepdim=True)
            return grid_counter_out, grid_semantic_out
    # if there is no semantic map, aggregate hit counter without Bayes filter
    else:
        # if there is a semantic segmentation, warn the user and count hits without Bayes filter
        if point_cloud_logprobs is not None:
            if verbose:
                print('Warning: semantic segmentation for point cloud detected, but no semantic map given! Incrementing hit counter without Bayes filter.')
        return point_cloud_fusion_util_cuda.point_cloud_hit_counter_free_function(grid_counter, point_cloud_locations, min_range, max_range)


import os

import torch
from torch.utils.cpp_extension import load

parent_dir = os.path.dirname(os.path.abspath(__file__))

counting_model_util_cuda = load(
        name='counting_model_util_cuda',
        sources=[
            os.path.join(parent_dir, path)
            for path in ['cuda/counting_model.cpp', 'cuda/counting_model.cu']],
        verbose=True)

def apply_counting_model(grid_counter, ray_origins, ray_directions, ray_distances, min_range, max_range, grid_semantic=None, ray_semseg=None, n_steps=2048, background_range = 5., verbose=False, invalidate_background=False, assert_inputs=True):
    if assert_inputs:
        assert grid_counter.is_cuda
        assert ray_origins.is_cuda
        assert ray_directions.is_cuda
        assert ray_distances.is_cuda
        assert grid_semantic is None or (grid_semantic.is_cuda and not (grid_semantic >= 0.).any())
        assert ray_semseg is None or (ray_semseg.is_cuda and not (ray_semseg >= 0.).any())
        assert min_range.is_cuda
        assert max_range.is_cuda
    # TODO: assert validity of inputs and document function
    # make sure the shape of grid_counter is suitable
    if len(grid_counter.shape) == 4:
        if verbose:
            print('Warning: grid_counter should be 5D, but is 4D. Adding dummy batch dimension.')
        grid_counter = grid_counter.unsqueeze(0)
    elif len(grid_counter.shape) == 3:
        if verbose:
            print('Warning: grid_counter should be 5D, but is 3D. Adding dummy batch and channel dimensions.')
        grid_counter = grid_counter.unsqueeze(0).unsqueeze(1)
    # check whether there is a channel for a miss counter - if not assume that only hits are counted
    n_counters = grid_counter.shape[1]
    # make sure the shapes of min_range and max_range are suitable
    if len(min_range.shape) == 1:
        if verbose:
            print('Warning: min_range should be 2D, but is not. Adding dummy batch dimension.')
        min_range = min_range.unsqueeze(0)
    if min_range.shape[0] != grid_counter.shape[0]:
        if verbose:
            print('Warning: min_range should have the same batch dimension as grid_counter, but does not. Assuming equal ranges for all scenes in batch.')
        min_range = min_range.expand(grid_counter.shape[0], -1)
    if len(max_range.shape) == 1:
        if verbose:
            print('Warning: max_range should be 2D, but is not. Adding dummy batch dimension.')
        max_range = max_range.unsqueeze(0)
    if max_range.shape[0] != grid_counter.shape[0]:
        if verbose:
            print('Warning: max_range should have the same batch dimension as grid_counter, but does not. Assuming equal ranges for all scenes in batch.')
        max_range = max_range.expand(grid_counter.shape[0], -1)
    # if provided, make sure the shape of the semantic voxel grid is suitable
    if grid_semantic is not None:
        if len(grid_semantic.shape) == 4:
            if verbose:
                print('Warning: grid_semantic should be 5D, but is 4D. Adding dummy batch dimension.')
            grid_semantic = grid_semantic.unsqueeze(0)
        # compute number of class channels
        n_classes = grid_semantic.shape[1]
    else:
        n_classes = 0
    # if only hits need to be counted, refer to the corresponding functions
    if n_counters == 1:
        # check whether a semantic map was provided
        if n_classes > 0:
            # if there is no semantic segmentation for rays, warn the user and apply counting model without it
            if ray_semseg is None:
                if verbose:
                    print('Warning: semantic voxel grid detected, but no semantic segmentation provided for rays! Counting hits without Bayes filter.')
                return counting_model_util_cuda.hit_counter_free_function(grid_counter, ray_origins, ray_directions, ray_distances, min_range, max_range)
            else:
            # if there is a semantic segmentation, make sure that the number of classes matches
                assert ray_semseg.shape[-1] == n_classes
                # modify distances for background class rays
                ray_distances_background = ray_distances.clone()
                ray_distances_background[torch.isnan(ray_semseg).any(-1)] = float('nan') if invalidate_background else background_range
                # obtain aggregated hit counter and fused semantic map
                grid_counter_out, grid_semantic_out = counting_model_util_cuda.hit_counter_bayes_free_function(grid_counter, grid_semantic, ray_origins, ray_directions, ray_distances_background, ray_semseg, min_range, max_range)
                # normalize semantic map
                grid_semantic_out -= torch.logsumexp(grid_semantic_out, dim=1, keepdim=True)
                return grid_counter_out, grid_semantic_out
        # if there is no semantic map, aggregate hit counter without Bayes filter
        else:
            # if there is a semantic segmentation, warn the user, mask background and count hits without Bayes filter
            if ray_semseg is not None:
                ray_distances_background = ray_distances.clone()
                # we never want to count background rays as hits
                #ray_distances_background[ray_semseg.sum(-1) > 0] = -1.
                ray_distances_background[torch.isnan(ray_semseg).any(-1)] = float('nan')
                if verbose:
                    print('Warning: semantic segmentation along rays detected, but no semantic map given! Incrementing hit counter for foreground pixels without Bayes filter.')
            else:
                ray_distances_background = ray_distances
            return counting_model_util_cuda.hit_counter_free_function(grid_counter, ray_origins, ray_directions, ray_distances_background, min_range, max_range)
    # on the other hand, if misses are counted too, refer to the corresponding functions
    else:
        if n_counters > 2:
            if verbose:
                print('Warning: grid_counter should have 1 or 2 channels, but has more. Only using first two channels.')
        # check whether a semantic map was provided
        if n_classes > 0:
            # if there is no semantic segmentation, warn the user and apply counting model without it
            if ray_semseg is None:
                if verbose:
                    print('Warning: semantic voxel grid detected, but no semantic segmentation provided for rays! Applying counting model without Bayes filter.')
                return counting_model_util_cuda.counting_model_free_function(grid_counter, ray_origins, ray_directions, ray_distances, min_range, max_range, n_steps)
            else:
                # if there is a semantic segmentation, make sure that the number of classes matches
                assert ray_semseg.shape[-1] == n_classes
                # modify distances for background class rays
                ray_distances_background = ray_distances.clone()
                ray_distances_background[torch.isnan(ray_semseg).any(-1)] = float('nan') if invalidate_background else background_range
                # apply counting model and obtain fused semantic map
                grid_counter_out, grid_semantic_out = counting_model_util_cuda.counting_model_bayes_free_function(grid_counter, grid_semantic, ray_origins, ray_directions, ray_distances_background, ray_semseg, min_range, max_range, n_steps)
                # normalize semantic map
                grid_semantic_out -= torch.logsumexp(grid_semantic_out, dim=1, keepdim=True)
                return grid_counter_out, grid_semantic_out
        # if there are no class channels, apply counting model without Bayes filter
        else:
            # if there is a semantic segmentation, warn the user and apply counting model without using it
            if ray_semseg is not None:
                ray_distances_background = ray_distances.clone()
                ray_distances_background[torch.isnan(ray_semseg).any(-1)] = float('nan') if invalidate_background else background_range
                if verbose:
                    print('Warning: semantic segmentation along rays detected, but no semantic map given! Applying counting model without Bayes filter.')
            else:
                ray_distances_background = ray_distances
            return counting_model_util_cuda.counting_model_free_function(grid_counter, ray_origins, ray_directions, ray_distances_background, min_range, max_range, n_steps)
